

from pathlib import Path
from typing import Any, Dict, List, Tuple
from dotenv import dotenv_values
import os 

class Config:
    def __init__(self, config_dict: Dict[str, Any]):
        self._config_dict = config_dict

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_config_dict":
            super().__setattr__(name, value)
        else:
            self._config_dict[name] = value

    def __getattr__(self, name: str) -> Any:
        try:
            return self._config_dict[name]
        except KeyError:
            raise AttributeError(f"'Config' object has no attribute '{name}'")

    def __getitem__(self, key: str) -> Any:
        return self._config_dict[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._config_dict[key] = value

    def __delitem__(self, key: str) -> None:
        del self._config_dict[key]

    def __contains__(self, key: str) -> bool:
        return key in self._config_dict

    def get(self, key: str, default: Any = None) -> Any:
        return self._config_dict.get(key, default)

    def keys(self) -> List[str]:
        return list(self._config_dict.keys())

    def values(self) -> List[Any]:
        return list(self._config_dict.values())

    def items(self) -> List[Tuple[str, Any]]:
        return list(self._config_dict.items())

def walk_up(d, f=None) -> List[Path]:
    d = Path(d).resolve()
    paths = []
    while True:
        d = d.parent
        if d == Path('/'):
            break

        if f is not None:
            paths.append(d / f)
        else:
            paths.append(d)

    return paths

def get_config_dirs(cwd=None, root=Path('/'), home=Path().home()) -> List[Path]:
    
    if cwd is None:
        cwd = Path.cwd()
    else:
        cwd = Path(cwd)

    root = Path(root)
    home = Path(home)

    return  [
        cwd,
        home.joinpath('.jtl'),
        root / Path('etc/jtl'), 
        root / Path('etc/jtl/secrets'), 
        root / Path('app/config'),
        root / Path('app/secrets'),
        cwd.joinpath('secrets'),
        cwd.parent.joinpath('secrets'),
    ] 


def find_config_files(file: str | List[str], dirs: List[str] | List[Path] = None) -> Path:
    """Find the first instance of a config file, from  a list of possible files, 
    in a list of directories. Return the first file that exists. """

    if isinstance(file, str):
        file = [file]

    if dirs is None:
        dirs = get_config_dirs()

    files = []
    for d in dirs:
        for f in file:
            p = Path(d) / f 
            if p.exists():
                files.append(p)  
    
    if files:
        return files
    else:
        raise FileNotFoundError(f"Could not find any of {file} in {dirs}")



def find_config_file(file: str | List[str], dirs: List[str] | List[Path] = None) -> Path:
    files = find_config_files(file, dirs)
    if len(files) > 1:
        raise FileNotFoundError(f"Found multiple files: {files}")
    return files[0]


def get_config(file: str | Path | List[str] | List[Path] = None, 
               dirs: List[str] | List[Path] = None) -> Config:

    """ Get the first config file found. The is for when you 
    just want one config file, but there may be multiple places for it. """

    if file is None:
        file = 'config.env'


    if '/' in str(file):
        fp = Path(file)
    else:
        fp = find_config_file(file, dirs)

    config = {
        '__CONFIG_PATH': str(fp.absolute()),
        **os.environ,
        **dotenv_values(fp),
    }

    return Config(config)

def get_config_tree(config_root: Path, deploy_name='devel') -> Config:
    """Assemble a configuration from a tree of config files. In this case
    (* different from get_config) the config is split into multiple 
    parts, and the parts are assembled into a single config object, and all of the
    configurations are stored in a set of subdirs of a common root directory.
    
    For each of the dirs 'config' and secret', the function will look for a file
    names 'config.env' and then '{deploy_name}.env' ( either 'devel' or 'prod' ) 
    and combine them into a single config object.
    
    So, if they exist, these files will be read and combined: 
    
        'config/config.env',
        '{deploy_name}.env',
        'secrets/config.env',
        'secrets/{deploy_name}.env',
    
    """
    
    root = Path(config_root).resolve()
    
    tree = [
        'config/config.env',
        'config/{deploy_name}.env',
        'secrets/secret.env',
        'secrets/{deploy_name}.env',
    ]
    
    d = {}
    configs = []
    
    for e in tree:
        f =  root / Path(e.format(deploy_name=deploy_name))
      
        if f.exists():
            d.update(dotenv_values(f))
            configs.append(f)
            
    d.update(os.environ)
    d['__CONFIG_PATH'] = configs
    
    return Config(d)
            
        
    
    
    

def path_interp(path: str, **kwargs) -> Tuple[str, Dict[str, Any]]:
    """
    Interpolates the parameters into the endpoint URL. So if you have a path
    like '/api/v1/leagues/:league_id/teams/:team_id' and you call

            path_interp(path, league_id=1, team_id=2, foobar=3)

    it will return '/api/v1/leagues/1/teams/2', along with a dictionary of
    the remaining parameters {'foobar': 3}.

    :param path: The endpoint URL template with placeholders.
    :param kwargs: The keyword arguments where the key is the placeholder (without ':') and the value is the actual value to interpolate.

    :return: A string with the placeholders in the path replaced with actual values from kwargs.
    """

    params = {}
    for key, value in kwargs.items():
        placeholder = f":{key}"  # Placeholder format in the path
        if placeholder in path:
            path = path.replace(placeholder, str(value))
        else:
            # Remove the trailing underscore from the key, so we can use params
            # like 'from' that are python keywords.
            params[key.rstrip('_')] = value

    return path, params



from datetime import datetime, timedelta
from src.jtlutil.datecalc import start_of_today, start_of_yesterday
from src.jtlutil.datecalc import end_of_today


def test_start_of_today():
    result = start_of_today()
    now = datetime.now()
    expected = datetime(now.year, now.month, now.day)
    assert result == expected, f"Expected {expected}, but got {result}"

def test_start_of_yesterday():
    result = start_of_yesterday()
    now = datetime.now()
    expected = datetime(now.year, now.month, now.day) - timedelta(days=1)
    assert result == expected, f"Expected {expected}, but got {result}"

def test_end_of_today():
    result = end_of_today()
    now = datetime.now()
    expected = datetime(now.year, now.month, now.day) + timedelta(days=1, seconds=-1)
    assert result == expected, f"Expected {expected}, but got {result}"



import logging
import os
import secrets
import sqlite3
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

import sqlitedict
from dotenv import dotenv_values
from flask import Flask, current_app, session, g
from flask_login import LoginManager, UserMixin
from flask_session import Session
from jtlutil.config import get_config, get_config_tree


class User(UserMixin):
    """Represents a user with attributes fetched from Google OAuth."""
    def __init__(self, user_data):
        self.user_data = user_data
        self.id = user_data['id']
        self.primary_email = user_data['primaryEmail']
        self.groups = user_data.get('groups', [])
        self.org_unit = user_data.get('orgUnitPath', '')
        self._is_admin = user_data.get('isAdmin', False)

    @property
    def is_league(self):
        """Return true if the user is a League user."""
        return self.primary_email.endswith('@jointheleague.org')

    @property
    def is_student(self):
        """Return true if the user is a student."""
        return self.primary_email.endswith('@students.jointheleague.org')

    @property
    def is_admin(self):
        return self._is_admin and self.is_league

    @property
    def is_staff(self):
        return self.is_league and 'staff@jointheleague.org' in self.groups
        
    @property
    def role(self):
        if self.is_admin:
            return "admin"
        elif self.is_staff:
            return "staff"
        elif self.is_student:
            return "student"
        elif self.is_league:
            return "league"
        else:
            return "Public"
        
    @property
    def is_public(self):
        return not self.is_league
        
    def get_full_user_info(self):
        return self.user_data



def is_running_under_gunicorn():
    """Return true if the app is running under Gunicorn, 
    which implies it is in production"""
    
    return "gunicorn" in os.environ.get(
        "SERVER_SOFTWARE", ""
    ) or "gunicorn" in os.environ.get("GUNICORN_CMD_ARGS", "")


def get_payload(request) -> dict:
    """Get the payload from the request, either from the form or json"""
        
    if request.content_type == "application/json":
        payload = request.get_json()
    else:
        payload = request.form.to_dict()

    # add date/time in iso format
    payload["_created"] = datetime.now().isoformat()

    return payload


def init_logger(app):
    """Initialize the logger for the app, either production or debug"""
    if is_running_under_gunicorn():
        gunicorn_logger = logging.getLogger("gunicorn.error")
        app.logger.handlers = gunicorn_logger.handlers
        app.logger.setLevel(gunicorn_logger.level)
        app.logger.info("Logger initialized for gunicorn")
    else:
        logging.basicConfig(level=logging.INFO)
        app.logger.setLevel(logging.DEBUG)
        app.logger.info("Logger initialized for flask")


def configure_config(app):
        # Determine if we're running in production or development
    if is_running_under_gunicorn():
        config_file_name = "prod.env"
    else:
        config_file_name = "devel.env"
    
        # Bypass the HTTPS requirement, because we are either running in development, 
        # or behind a proxy than handles https. May be better to set the X-Forwarded-Proto, 
        # but that looks really complicated. 
        os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1' 

    # Load configuration
    config = get_config(config_file_name)
    
    # Set the Flask secret key
    app.secret_key = config["SECRET_KEY"]
    
    # Resolve the path to the secrets file
    if 'SECRETS_FILE_NAME' in config:
        config["SECRETS_FILE"] = (Path(config['__CONFIG_PATH']).parent / config['SECRETS_FILE_NAME']).resolve()

    # Store 
    
    app.app_config = config
    
    return config


def configure_config_tree(app):
        # Determine if we're running in production or development
    if is_running_under_gunicorn():
        deploy = "prod"
        config_dir = '/app'
    else:
        deploy = "devel"
        config_dir = Path().cwd()
    
        # Bypass the HTTPS requirement, because we are either running in development, 
        # or behind a proxy than handles https. May be better to set the X-Forwarded-Proto, 
        # but that looks really complicated. 
        os.environ['OAUTHLIB_INSECURE_TRANSPORT'] = '1' 

    app.logger.info(f"Configuring for deployment: {deploy}")
    config = get_config_tree(config_dir, deploy_name=deploy)

    # Set the Flask secret key
    app.secret_key = config["SECRET_KEY"]
    
    # Resolve the path to the secrets file
    if 'SECRETS_FILE_NAME' in config:
        config["SECRETS_FILE"] = (Path(config['__CONFIG_PATH']).parent / config['SECRETS_FILE_NAME']).resolve()

    # Store 
    
    app.app_config = config
    
    return config

def configure_app_dir(app):
    
      # Configure the appdir

    app.app_config.app_dir = app_dir = Path(app.app_config.APP_DIR)
    
    if not app_dir.exists():
        app_dir.mkdir(parents=True)
        
    app.app_config.data_dir = data_dir = Path(app.app_config.DATA_DIR) 
        
    app.app_config.db_dir = db_dir = app.app_config.data_dir / 'db'
    
    if not db_dir.exists():
        db_dir.mkdir(parents=True)
        
   
    return app_dir, db_dir

def setup_sqlite_sessions(app, session_expire_time=60*60*24*1): 
        # Setup sessions
        
    db_path = app.app_config.db_dir / "sessions.db"
    app.config['SESSION_TYPE'] = 'sqlalchemy'
    app.config['SQLALCHEMY_DATABASE_URI'] = f'sqlite:///{db_path}'
    Session(app) # Initialize the session
    
    # Set session expiration time
    #app.config['PERMANENT_SESSION_LIFETIME'] = session_expire_time
    app.config['SESSION_CLEANUP_N_REQUESTS'] = 100
    app.config['SESSION_SERIALIZATION_FORMAT'] = 'json'
    app.config['SESSION_COOKIE_SECURE'] = True
    
    
def insert_query_arg(url, key, value):
    parsed_url = urlparse(url)
    query_params = parse_qs(parsed_url.query)
    query_params[key] = value
    new_query_string = urlencode(query_params, doseq=True)
    return urlunparse(parsed_url._replace(query=new_query_string))





import uuid

import requests
from flask import (current_app, flash, get_flashed_messages, 
                   redirect, render_template, request, session, url_for, g)
from flask import Blueprint
from flask_login import (current_user, login_user, logout_user, LoginManager)
from jtlutil.flask.flaskapp import insert_query_arg
from jtlutil.jwtencode import decrypt_token, encrypt_token
import logging
from .flaskapp import User

auth_bp = Blueprint("auth", __name__)

logger = logging.getLogger(__name__)

login_manager = LoginManager()
login_manager.init_app(auth_bp)
login_manager.login_view = 'auth.login'
login_manager = auth_bp.login_manager = login_manager

@auth_bp.login_manager.user_loader
def load_lm_user(user_id):
    current_app.logger.info(f"Request to load  user from login manager with ID: {user_id}")

    if "user" in session:
        current_app.logger.info(f"User data already in session")
        user_data = session["user"]
        return User(user_data)

    return None


def get_session_user(app,session_id):
    
    enc_key = bytes.fromhex(current_app.app_config['ENCRYPTION_KEY'])
    
    t = encrypt_token(session_id, enc_key)
    
    auth_server_url = current_app.app_config["AUTH_SERVER_URL"]
    user_endpoint = f"{auth_server_url}/user"
    params = {"ssoid": t}
    
    response = requests.get(user_endpoint, params=params)
    
    if response.status_code == 200:
        user_data = response.json()
        
        return user_data
    else:
        current_app.logger.error(f"Failed to fetch user data: {response.status_code} {response.text}")
        return None

def load_user(app):
    """When the authentication server redirects back to the app here, it will include a query
    parameter `ssoid` which is the encrypted session ID. This function decrypts
    the session ID and loads the user data into the session. We can look up this
    session id in the cache and load the user data into the session.
    """

    query_args = request.args.to_dict()
    ssoid = query_args.get("ssoid")
    
    current_app.logger.info(f"Loading user with ssoid: {ssoid}")
    
    if ssoid:
        # Decrypt it
        session_id = decrypt_token(
            ssoid, bytes.fromhex(app.app_config["ENCRYPTION_KEY"])
        )
        user_data = get_session_user(app, session_id)
        session["user"] = user_data
        
        user = User(user_data)

        current_app.logger.info(f"Logging in user: {user.primary_email}")

        login_user(user)
    else:
        current_app.logger.error("No ssoid in query args")
        #assert False, "No ssoid in query args"
        


@auth_bp.route("/login")
def login():

    next = request.args.get("next", url_for("index", _external=True))
    
    login_url = insert_query_arg(
        current_app.app_config["AUTH_SERVER_URL"] + "/login",
        "redirect", next,
    )
    current_app.logger.info(f"Redirecting to login server at {login_url}")

    get_flashed_messages()

    return redirect(login_url)


@auth_bp.route("/logout")
def logout():

    next = request.args.get("next", url_for("index", _external=True))

    if current_user.is_authenticated:
        current_app.logger.info(f"User {current_user.id} logging out")

    #uncache_user(session["session_id"])
    session.clear()
    logout_user()
    current_app.logger.info(f"ser logged out")

    flash("You have been logged out.", "info")
    return redirect(next)

